@@ -109,6 +109,22 @@ impl Error {
             Self::Code(code) => unsafe { s2n_error_get_type(*code) as _ },
         }
     }
+
+    pub fn alert(&self) -> Option<u8> {
+        match self {
+            Self::InvalidInput => None,
+            Self::Code(_code) => {
+                None
+                // TODO: We should use the new s2n-tls method
+                //       once it's available.
+                // let mut alert = 0;
+                // match call!(s2n_error_get_alert(*code, &mut alert)) {
+                //     Ok(_) => Some(alert),
+                //     Err(_) => None,
+                // }
+            }
+        }
+    }
 }
 
 /// # Safety
