@@ -13,7 +13,7 @@
  * permissions and limitations under the License.
  */
 
-#include <s2n.h>
+#include "api/s2n.h"
 
 #include "crypto/s2n_hash.h"
 #include "crypto/s2n_signature.h"
